import { HttpClient, HttpHeaders, HttpParams } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Observable } from 'rxjs/internal/Observable';
import { catchError } from 'rxjs/operators';

import { Account } from './account';
import { AccountBalance } from './account-balance';
import { ErrorLoggerService } from './error-logger.service';

const url = '/api/accounts';
const serviceName = 'AccountService';

@Injectable({ providedIn: 'root' })
export class AccountService {
  constructor(private http: HttpClient, private log: ErrorLoggerService) {}

  get(id: string | null): Observable<Account> {
    const getUrl: string = id === null ? `${url}` : `${url}/${id}`;
    return this.http
      .get<Account>(getUrl)
      .pipe(catchError(this.log.handleError(serviceName, `get id=${id}`))) as Observable<Account>;
  }

  list(): Observable<Account[]> {
    return this.http
      .get<Account[]>(`${url}/list`)
      .pipe(catchError(this.log.handleError(serviceName, 'list'))) as Observable<Account[]>;
  }

  paymentAutocomplete(query: string): Observable<Account[]> {
    const options = { params: new HttpParams().set('q', query).set('t', '1') };
    return this.http
      .get<Account[]>(`${url}/query`, options)
      .pipe(catchError(this.log.handleError(serviceName, 'list'))) as Observable<Account[]>;
  }

  receiptAutocomplete(query: string): Observable<Account[]> {
    const options = { params: new HttpParams().set('q', query).set('t', '1') };
    return this.http
      .get<Account[]>(`${url}/query`, options)
      .pipe(catchError(this.log.handleError(serviceName, 'list'))) as Observable<Account[]>;
  }

  save(account: Account): Observable<null> {
    return this.http
      .post<null>(`${url}`, account)
      .pipe(catchError(this.log.handleError(serviceName, 'save'))) as Observable<null>;
  }

  update(account: Account): Observable<null> {
    return this.http
      .put<null>(`${url}/${account.id}`, account)
      .pipe(catchError(this.log.handleError(serviceName, 'update'))) as Observable<null>;
  }

  saveOrUpdate(account: Account): Observable<null> {
    if (!account.id) {
      return this.save(account);
    }
    return this.update(account);
  }

  delete(id: string): Observable<Account> {
    return this.http
      .delete<Account>(`${url}/${id}`)
      .pipe(catchError(this.log.handleError(serviceName, 'delete'))) as Observable<Account>;
  }

  autocomplete(query: string): Observable<Account[]> {
    const options = { params: new HttpParams().set('q', query) };
    return this.http
      .get<Account[]>(`${url}/query`, options)
      .pipe(catchError(this.log.handleError(serviceName, 'autocomplete'))) as Observable<Account[]>;
  }

  balance(id: string, date: string): Observable<AccountBalance> {
    const options = { params: new HttpParams().set('d', date) };
    return this.http
      .get<AccountBalance>(`${url}/${id}/balance`, options)
      .pipe(catchError(this.log.handleError(serviceName, 'balance'))) as Observable<AccountBalance>;
  }
}
