import { Component } from '@angular/core';
import { FormBuilder, FormControl, FormGroup } from '@angular/forms';
import { MatDatepicker } from '@angular/material/datepicker';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import * as moment from 'moment';
import { Moment } from 'moment';

import { AuthService } from '../auth/auth.service';
import { ToasterService } from '../core/toaster.service';

import { EmployeeFunctionsService } from './employee-functions.service';

@Component({
  selector: 'app-employee-functions',
  templateUrl: './employee-functions.component.html',
  styleUrls: ['./employee-functions.component.css'],
})
export class EmployeeFunctionsComponent {
  creditSalaryForm: FormGroup;
  attendanceRecordForm: FormGroup;
  fingerprintForm: FormGroup;
  fingerprintFile: File | null = null;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private fb: FormBuilder,
    private dialog: MatDialog,
    private toaster: ToasterService,
    private auth: AuthService,
    private ser: EmployeeFunctionsService,
  ) {
    const startDate = moment().date(1);
    const finishDate = moment().date(startDate.daysInMonth());
    this.creditSalaryForm = this.fb.group({
      date: finishDate,
    });
    this.attendanceRecordForm = this.fb.group({
      startDate,
      finishDate,
    });
    this.fingerprintForm = this.fb.group({});
  }

  chosenYearHandler(normalizedYear: Moment) {
    const dateControl = this.creditSalaryForm.get('date') as FormControl;
    const ctrlValue = dateControl.value;
    ctrlValue.year(normalizedYear.year());
    ctrlValue.date(ctrlValue.daysInMonth());
    dateControl.setValue(ctrlValue);
  }

  chosenMonthHandler(normlizedMonth: Moment, datepicker: MatDatepicker<Moment>) {
    const dateControl = this.creditSalaryForm.get('date') as FormControl;
    const ctrlValue = dateControl.value;
    ctrlValue.month(normlizedMonth.month());
    ctrlValue.date(ctrlValue.daysInMonth());
    dateControl.setValue(ctrlValue);
    datepicker.close();
  }

  creditSalary() {
    const date = moment((this.creditSalaryForm.get('date') as FormControl).value).format(
      'DD-MMM-YYYY',
    );
    if (!date) {
      this.toaster.show('Danger', 'Please choose a valid date.');
      return;
    }
    this.ser.creditSalary(date).subscribe(
      () => {
        this.toaster.show('Success', 'Salaries Credited');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  attendanceRecordUrl() {
    const startDate = moment(
      (this.attendanceRecordForm.get('startDate') as FormControl).value,
    ).format('DD-MMM-YYYY');
    const finishDate = moment(
      (this.attendanceRecordForm.get('finishDate') as FormControl).value,
    ).format('DD-MMM-YYYY');
    if (!startDate || !finishDate) {
      // this.toaster.show('Danger', 'Please choose a start and finish date.');
      return '';
    }
    return `/attendance-report?s=${startDate}&f=${finishDate}`;
  }

  fingerPrintUrl() {
    const startDate = moment(
      (this.attendanceRecordForm.get('startDate') as FormControl).value,
    ).format('DD-MMM-YYYY');
    const finishDate = moment(
      (this.attendanceRecordForm.get('finishDate') as FormControl).value,
    ).format('DD-MMM-YYYY');
    if (!startDate || !finishDate) {
      // this.toaster.show('Danger', 'Please choose a start and finish date.');
      return '';
    }
    return `/fingerprint-report?s=${startDate}&f=${finishDate}`;
  }

  detectFiles(event: Event) {
    // eslint-disable-next-line prefer-destructuring
    this.fingerprintFile = ((event.target as HTMLInputElement).files as FileList)[0];
  }

  uploadFingerprints() {
    if (!this.fingerprintFile) {
      this.toaster.show('Danger', 'Please choose a file first!');
      return;
    }
    this.ser.uploadFingerprints(this.fingerprintFile).subscribe(
      () => {
        this.toaster.show('Success', 'Fingerprints uploaded');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }
}
