import { Component, OnInit } from '@angular/core';
import { FormArray, FormBuilder, FormControl, FormGroup } from '@angular/forms';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import * as moment from 'moment';
import { BehaviorSubject } from 'rxjs';

import { environment } from '../../environments/environment';
import { AuthService } from '../auth/auth.service';
import { AccountType } from '../core/account-type';
import { ToasterService } from '../core/toaster.service';
import { ConfirmDialogComponent } from '../shared/confirm-dialog/confirm-dialog.component';

import { LockDataSource } from './lock-datasource';
import { LockInfo } from './lock-info';
import { SettingsService } from './settings.service';

@Component({
  selector: 'app-settings',
  templateUrl: './settings.component.html',
  styleUrls: ['./settings.component.css'],
})
export class SettingsComponent implements OnInit {
  public lockObservable = new BehaviorSubject<LockInfo[]>([]);
  dataSource: LockDataSource = new LockDataSource(this.lockObservable);

  accountTypes: AccountType[] = [];
  voucherTypes: AccountType[] = [];

  lockInfoForm: FormGroup;
  lockInformation: LockInfo[];
  lockOlder: boolean = false;
  olderRolling: boolean = false;
  lockNewer: boolean = false;
  newerRolling: boolean = false;

  rebaseDataForm: FormGroup;

  maintenance: { enabled: boolean; user: string };

  displayedColumns = ['index', 'validity', 'voucherTypes', 'accountTypes', 'lock', 'action'];

  version: string;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private fb: FormBuilder,
    private dialog: MatDialog,
    private toaster: ToasterService,
    public auth: AuthService,
    private ser: SettingsService,
  ) {
    const startDate = moment().date(1);
    const finishDate = moment().date(startDate.daysInMonth());
    this.lockInfoForm = this.fb.group({
      validFrom: '',
      validTill: '',
      index: 0,
      lockOlder: '',
      olderRolling: '',
      olderDate: finishDate,
      olderDays: 0,
      lockNewer: '',
      newerRolling: '',
      newerDate: finishDate,
      newerDays: 0,
      accountTypes: this.fb.array([]),
      voucherTypes: this.fb.array([]),
    });
    this.lockInformation = [];
    this.rebaseDataForm = this.fb.group({
      date: moment(),
    });

    this.listenToLockForm();
    this.maintenance = { enabled: false, user: '' };
    this.version = environment.version;
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as {
        maintenance: { enabled: boolean; user: string };
        lockInformation: [LockInfo];
        accountTypes: AccountType[];
        voucherTypes: AccountType[];
      };
      this.accountTypes = data.accountTypes;
      this.voucherTypes = data.voucherTypes;
      this.showLockInformation(data.lockInformation);
      this.maintenance = data.maintenance;
    });
  }

  showLockInformation(info: LockInfo[]) {
    this.lockInformation = info;
    this.lockInfoForm.setControl(
      'accountTypes',
      this.fb.array(
        this.accountTypes.map((x) =>
          this.fb.group({
            accountType: false,
          }),
        ),
      ),
    );
    this.lockInfoForm.setControl(
      'voucherTypes',
      this.fb.array(
        this.voucherTypes.map((x) =>
          this.fb.group({
            voucherType: false,
          }),
        ),
      ),
    );
    this.dataSource = new LockDataSource(this.lockObservable);
    this.lockObservable.next(this.lockInformation);
  }

  listenToLockForm() {
    (this.lockInfoForm.get('lockOlder') as FormControl).valueChanges.subscribe((x) => {
      this.lockOlder = x;
    });
    (this.lockInfoForm.get('lockNewer') as FormControl).valueChanges.subscribe((x) => {
      this.lockNewer = x;
    });
    (this.lockInfoForm.get('olderRolling') as FormControl).valueChanges.subscribe((x) => {
      this.olderRolling = x;
    });
    (this.lockInfoForm.get('newerRolling') as FormControl).valueChanges.subscribe((x) => {
      this.newerRolling = x;
    });
  }

  saveLock() {
    const item = new LockInfo();
    if (this.lockOlder && this.olderRolling) {
      item.start.days = +(this.lockInfoForm.get('olderDays') as FormControl).value;
    } else if (this.lockOlder && !this.olderRolling) {
      item.start.date = moment((this.lockInfoForm.get('olderDate') as FormControl).value).format(
        'DD-MMM-YYYY',
      );
    }
    if (this.lockNewer && this.newerRolling) {
      item.finish.days = +(this.lockInfoForm.get('newerDays') as FormControl).value;
    }
    if (this.lockNewer && !this.newerRolling) {
      item.finish.date = moment((this.lockInfoForm.get('newerDate') as FormControl).value).format(
        'DD-MMM-YYYY',
      );
    }
    const atArray = this.lockInfoForm.get('accountTypes') as FormArray;
    this.accountTypes.forEach((at, index) => {
      if (atArray.controls[index].value.accountType) {
        item.accountTypes.push({ id: at.id, name: at.name });
      }
    });

    const vtArray = this.lockInfoForm.get('voucherTypes') as FormArray;
    this.voucherTypes.forEach((vt, index) => {
      if (vtArray.controls[index].value.voucherType) {
        item.voucherTypes.push({ id: vt.id, name: vt.name });
      }
    });

    const validFrom = (this.lockInfoForm.get('validFrom') as FormControl).value;
    if (validFrom) {
      item.validFrom = moment(validFrom).format('DD-MMM-YYYY');
    }

    const validTill = (this.lockInfoForm.get('validTill') as FormControl).value;
    if (validTill) {
      item.validTill = moment(validTill).format('DD-MMM-YYYY');
    }
    item.index = +(this.lockInfoForm.get('index') as FormControl).value;

    this.ser.setLockInformation(item).subscribe(
      (result) => {
        this.showLockInformation(result);
        this.toaster.show('Success', 'Lock information Updated');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  deleteLock(row: LockInfo) {
    this.ser.deleteLockInformation(row.id as string).subscribe((x) => {
      this.showLockInformation(x);
    });
  }

  confirmRebase(): void {
    const rebaseDate = moment((this.rebaseDataForm.get('date') as FormControl).value).format(
      'DD-MMM-YYYY',
    );
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: {
        title: 'Rebase the database?',
        content: `Are you sure you want to rebase the database as on ${rebaseDate}? This action is destructive and cannot be undone.`,
      },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.rebaseData(rebaseDate);
      }
    });
  }

  rebaseData(rebaseDate: string) {
    this.ser.rebaseDatabase(rebaseDate).subscribe(
      () => {
        this.toaster.show('Success', 'Data has been rebased!');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  toggleMaintenance() {
    this.ser.setMaintenance(!this.maintenance.enabled).subscribe((x) => {
      this.maintenance = x;
    });
  }
}
